//
// connection.cpp
// ~~~~~~~~~~~~~~
//
// Copyright (c) 2003-2024 Christopher M. Kohlhoff (chris at kohlhoff dot com)
//
// Distributed under the Boost Software License, Version 1.0. (See accompanying
// file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
//

#include "connection.hpp"
#include <utility>
#include <boost/log/trivial.hpp>
#include <boost/beast.hpp>

#include "connection_manager.hpp"


namespace http::server {
    connection::connection(boost::asio::ip::tcp::socket socket,
                           connection_manager &manager, request_handler handler)
        : socket_(std::move(socket)),
          connection_manager_(manager),
          request_handler_(std::move(handler)), request_(), reply_() {
    }

    void connection::start() {
        do_read();
    }

    void connection::stop() {
        socket_.close();
    }

    void connection::do_read() {
        request_parser_.reset();
        request_.headers.clear();
        request_.method.clear();
        request_.uri.clear();

        auto self(shared_from_this());
        socket_.async_read_some(boost::asio::buffer(buffer_), [this, self](boost::system::error_code ec, std::size_t bytes_transferred) {
            if (!ec) {
                request_parser::result_type result;
                std::tie(result, std::ignore) = request_parser_.parse(
                    request_, buffer_.data(), buffer_.data() + bytes_transferred);

                if (result == request_parser::good) {
                    request_handler_(request_, reply_);
                    do_write();
                } else if (result == request_parser::bad) {
                    stock_reply(bad_request, reply_);
                    do_write();
                } else {
                    do_read();
                }
            } else {
                connection_manager_.stop(shared_from_this());
            }
        });
    }

    void connection::do_write() {
        reply_.headers.push_back({.name = "Server", .value = "TerminalWebServer v0.1"});
        if (!reply_.content.empty()) {
            reply_.headers.push_back({.name = "Content-Length", .value = std::to_string(reply_.content.size())});
        }
        if (request_.http_version_major == 1) {
            reply_.headers.push_back({.name = "Connection", .value = "keep-alive"});
        }

        BOOST_LOG_TRIVIAL(info) << "HTTP query " << reply_.status << " " << request_.method << " " << request_.uri;

        auto self(shared_from_this());
        async_write(socket_, reply_.to_buffers(), [this, self](boost::system::error_code ec, std::size_t) {
            if (!ec) {
                // keep alive connection
                do_read();
            } else {
                connection_manager_.stop(shared_from_this());
            }
        });
    }
} // namespace http::server
